import { SlashCommandBuilder } from "@discordjs/builders";
import { Player, SearchResult, useMainPlayer, useQueue } from "discord-player";
import {
  AutocompleteInteraction,
  ChatInputCommandInteraction,
  Client,
  EmbedBuilder,
  GuildResolvable,
  VoiceBasedChannel,
} from "discord.js";
import { getLocale, getLocalizations } from "../../utils/locales";
import { Metadata } from "../../utils/metadata";
import { getFilename } from "../../utils/misc";

export default {
  scope: () => [],

  data: (client: Client) => {
    const filename = getFilename(__filename);
    const loc_default = client.locales.get(client.config.default_lang);
    if (!loc_default) {
      return;
    }

    return (
      new SlashCommandBuilder()
        .setName(filename.toLowerCase())
        .setDescription(loc_default.get(`c_${filename}_desc`) ?? "")
        .setNameLocalizations(getLocalizations(client, `c_${filename}_name`, true))
        .setDescriptionLocalizations(getLocalizations(client, `c_${filename}_desc`))

        // Command option
        .addStringOption((option) =>
          option
            .setName(loc_default.get(`c_${filename}_opt1_name`)?.toLowerCase() ?? "")
            .setDescription(loc_default.get(`c_${filename}_opt1_desc`) ?? "")
            .setNameLocalizations(getLocalizations(client, `c_${filename}_opt1_name`, true))
            .setDescriptionLocalizations(getLocalizations(client, `c_${filename}_opt1_desc`))
            .setAutocomplete(true),
        )
    );
  },

  interaction: async (interaction: ChatInputCommandInteraction, client: Client) => {
    const loc = getLocale(client, interaction.locale);
    const loc_default = client.locales.get(client.config.default_lang);
    const filename = getFilename(__filename);

    const member = client.guilds.cache
      .get(interaction.guildId ?? "")
      ?.members.cache.get(interaction.member?.user.id ?? "");

    if (!member?.voice.channelId) {
      return await interaction.reply({
        content: `❌ | ${loc.get("c_play1")}`,
        ephemeral: true,
      });
    }

    if (
      interaction.guild?.members.me?.voice.channelId &&
      member.voice.channelId !== interaction.guild.members.me.voice.channelId
    ) {
      return await interaction.reply({
        content: `❌ | ${loc.get("c_play2")} ${interaction.guild.members.me.voice.channel}`,
        ephemeral: true,
      });
    }

    const query = interaction.options.getString(
      loc_default?.get(`c_${filename}_opt1_name`) as string,
    );

    const player = useMainPlayer() as Player;
    if (!query) {
      // Now playing

      const queue = useQueue(interaction.guildId ?? "");

      if (queue) {
        const track = queue.history.currentTrack;
        if (track) {
          const embed = new EmbedBuilder()
            .setDescription(
              `${queue.node.createProgressBar()}\n\n${loc.get("c_play8")} ${track.requestedBy}`,
            )
            .setTitle(track.title + " • " + track.author)
            .setURL(track.url)
            .setThumbnail(track.thumbnail)
            .setFooter({ text: loc.get("c_play7") })
            .setTimestamp();
          return await interaction.reply({ embeds: [embed] });
        }
      }

      const embed = new EmbedBuilder().setDescription(`❌ | ${loc.get("c_play6")}`);
      return await interaction.reply({ embeds: [embed] });
    }

    const queue = player.nodes.create(interaction.guild as GuildResolvable, {
      defaultFFmpegFilters: ["silenceremove"],
      metadata: {
        channel: interaction.channel,
      } as Metadata,
    });

    // Verify vc connection
    try {
      if (!queue.connection) await queue.connect(member.voice.channel as VoiceBasedChannel);
    } catch {
      queue.delete();
      return await interaction.reply({
        content: `❌ | ${loc.get("c_play3")}`,
        ephemeral: true,
      });
    }

    await interaction.deferReply();
    const result = await player
      .search(query, {
        requestedBy: interaction.user,
      })
      .then((x) => x);

    if (result.isEmpty()) {
      const embed = new EmbedBuilder().setDescription(`❌ | \`${query}\` ${loc.get("c_play4")}.`);
      return await interaction.followUp({ embeds: [embed] });
    }

    let title;
    if (result.playlist) {
      queue.addTrack(result.playlist.tracks);
      title = result.playlist.title;
    } else {
      const track = result.tracks[0];

      queue.addTrack(track);
      title = track.title;
    }

    if (!queue.node.isPlaying()) {
      queue.node.play();
    }

    // TODO: When added to an existing queue (size of queue > 0):
    //   - Add position in queue
    //   - Add estimated time until playing
    return await interaction.followUp({
      content: `⏱️ | \`${title}\` ${loc.get("c_play5")}.`,
    });
  },

  autocomplete: async (interaction: AutocompleteInteraction) => {
    const loc = getLocale(interaction.client, interaction.locale);
    const loc_default = interaction.client.locales.get(interaction.client.config.default_lang);
    const filename = getFilename(__filename);

    const player = useMainPlayer() as Player;
    const query = interaction.options.getString(
      loc_default?.get(`c_${filename}_opt1_name`) as string,
      true,
    );

    const limit_value_discord = 100;

    if (query) {
      /* Since Discord wanna receive a response within 3 secs and results is async
       * and can take longer than that, exception of type 'Unknown interaction' (10062)
       * happens. */

      let timeoutId: NodeJS.Timeout;
      const delay = new Promise(function (_, reject) {
        timeoutId = setTimeout(function () {
          reject(new Error());
        }, 2800);
      });

      /* Create a race between a timeout and the search
       * At the end, Discord will always receive a response */
      let tracks = await Promise.race([
        delay,
        player.search(query, {
          requestedBy: interaction.user,
        }),
      ])
        .then((res) => {
          clearTimeout(timeoutId);
          return (res as SearchResult).tracks;
        })
        .catch(() => {
          return [];
        });

      // If tracks found
      if (tracks.length > 0) {
        if (tracks.length > 25) {
          tracks = tracks
            // Assure that URL is under the limit of Discord
            .filter((v) => v.url.length < limit_value_discord)
            // Slice the list if needed to the 25 first results
            .slice(0, 25);
        }

        // Returns a list of songs with their title and author
        return interaction.respond(
          tracks.map((t) => {
            let title = t.title;
            let author = t.author;
            let name = `${title} • ${author}`;

            // Slice returned data if needed to not exceed the length limit (100)
            if (name.length > 100) {
              const newTitle = title.substring(0, 40);
              if (title.length != newTitle.length) {
                title = `${newTitle}...`;
              }
              const newAuthor = author.substring(0, 40);
              if (author.length != newAuthor.length) {
                author = `${newAuthor}...`;
              }
              name = `${newTitle} • ${newAuthor}`;
            }

            return {
              name,
              value: t.url,
            };
          }),
        );
      }
    }
    return interaction.respond([
      { name: loc.get("c_play9"), value: query.slice(0, limit_value_discord) },
    ]);
  },
};
